# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2022/12/16 19:37
describe: 
"""
import requests
from czsc.fsa.base import request, FeishuApiBase
from czsc.fsa.spreed_sheets import SpreadSheets
from czsc.fsa.im import IM


def push_text(text: str, key: str) -> None:
    """使用自定义机器人推送文本消息到飞书群聊

    如何在群组中使用机器人: https://www.feishu.cn/hc/zh-CN/articles/360024984973

    :param text: 文本内容
    :param key: 机器人的key
    :return: None
    """
    api_send = f"https://open.feishu.cn/open-apis/bot/v2/hook/{key}"
    data = {"msg_type": "text", "content": {"text": text}}
    try:
        response = requests.post(api_send, json=data)
        assert response.json()['StatusMessage'] == 'success'
    except:
        print(f"{data} - 文本消息推送失败")




# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/12/13 17:48
describe: 验证信号计算的准确性，仅适用于缠论笔相关的信号，
          技术指标构建的信号，用这个工具检查不是那么方便
"""
import sys

sys.path.insert(0, '..')
import os
import numpy as np
from loguru import logger
from collections import OrderedDict
from czsc.data.ts_cache import TsDataCache
from czsc import CZSC, Signal
from czsc.traders.base import CzscTrader, check_signals_acc
from czsc.signals.tas import update_ma_cache
from czsc.utils import get_sub_elements, create_single_signal
from czsc import signals

os.environ['czsc_verbose'] = '1'


def bar_reversal_V230227(c: CZSC, di=1, avg_bp: int = 300, **kwargs) -> OrderedDict:
    """判断最近一根K线是否具有反转迹象

    **信号逻辑：**

    - 看多：当前K线为阴线，或阳线长上影; 且截止前一根K线，连续 3 / 5 / 8根K线累计涨幅超过 avg_bp * n，或 连续13根K线都是阳线
    - 反之，看空

    **信号列表：**

    - Signal('日线_D2A100_反转V230227_看多_任意_任意_0')
    - Signal('日线_D2A100_反转V230227_看空_任意_任意_0')

    :param c: CZSC对象
    :param di: 倒数第几根K线
    :param avg_bp: 平均单根K线的涨跌幅，用于判断是否是反转
    :return:
    """
    k1, k2, k3 = str(c.freq.value), f"D{di}A{avg_bp}", "反转V230227"
    v1 = "其他"
    _bars = get_sub_elements(c.bars_raw, di=di, n=14)

    if len(_bars) != 14:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    last_bar = _bars[-1]
    left_bar = _bars[:-1]

    # 阳线长上影
    last_bar_up_c1 = last_bar.close > last_bar.open and last_bar.upper > 2 * max(last_bar.solid, last_bar.lower)

    # 阴线长下影
    last_bar_dn_c1 = last_bar.close < last_bar.open and last_bar.lower > 2 * max(last_bar.solid, last_bar.upper)

    if last_bar.close < last_bar.open or last_bar_up_c1:
        # 连续3 / 5 / 8根K线累计涨幅超过 avg_bp * n / 10000
        up_c1 = (left_bar[-1].close / left_bar[-3].open - 1) / 3 > avg_bp / 10000
        up_c2 = (left_bar[-1].close / left_bar[-5].open - 1) / 5 > avg_bp / 10000
        up_c3 = (left_bar[-1].close / left_bar[-8].open - 1) / 8 > avg_bp / 10000

        # 连续13根K线都是阳线
        up_c4 = all(bar.close > bar.open for bar in left_bar)

        if any([up_c1, up_c2, up_c3, up_c4]):
            v1 = "看空"

    if last_bar.close > last_bar.open or last_bar_dn_c1:
        # 连续3 / 5 / 8根K线累计跌幅超过 avg_bp * n / 10000
        dn_c1 = (left_bar[-1].close / left_bar[-3].open - 1) / 3 < -avg_bp / 10000
        dn_c2 = (left_bar[-1].close / left_bar[-5].open - 1) / 5 < -avg_bp / 10000
        dn_c3 = (left_bar[-1].close / left_bar[-8].open - 1) / 8 < -avg_bp / 10000

        # 连续13根K线都是阴线
        dn_c4 = all(bar.close < bar.open for bar in left_bar)

        if any([dn_c1, dn_c2, dn_c3, dn_c4]):
            v1 = "看多"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def get_signals(cat: CzscTrader) -> OrderedDict:
    s = OrderedDict({"symbol": cat.symbol, "dt": cat.end_dt, "close": cat.latest_price})
    # 使用缓存来更新信号的方法
    s.update(bar_reversal_V230227(cat.kas['日线'], di=2, avg_bp=300))
    return s


if __name__ == '__main__':
    data_path = r'C:\ts_data'
    dc = TsDataCache(data_path, sdt='2010-01-01', edt='20211209')

    symbol = '000001.SZ'
    bars = dc.pro_bar_minutes(ts_code=symbol, asset='E', freq='15min',
                              sdt='20181101', edt='20210101', adj='qfq', raw_bar=True)

    check_signals_acc(bars, get_signals)



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2022/11/7 19:16
describe: 白仪 https://www.zhihu.com/people/bai-yi-520/posts 知乎上定义的一些信号

byi 是 bai yi 的缩写
"""
from collections import OrderedDict
from typing import List
from czsc import CZSC
from czsc.objects import Signal, BI, Direction
from czsc.utils import get_sub_elements
from czsc.utils.sig import is_symmetry_zs


def byi_symmetry_zs_V221107(c: CZSC, di=1):
    """对称中枢信号

    **信号逻辑：**

    从di笔往前数7/5/3笔，如果构成中枢，且所有笔的力度序列标
    准差小于均值的一定比例，则认为是对称中枢

    **信号列表：**
    
    - Signal('15分钟_D1B_对称中枢_否_向下_任意_0')
    - Signal('15分钟_D1B_对称中枢_是_向上_7笔_0')
    - Signal('15分钟_D1B_对称中枢_否_向上_任意_0')
    - Signal('15分钟_D1B_对称中枢_是_向下_3笔_0')
    - Signal('15分钟_D1B_对称中枢_是_向上_3笔_0')
    - Signal('15分钟_D1B_对称中枢_是_向下_5笔_0')
    - Signal('15分钟_D1B_对称中枢_是_向上_5笔_0')
    - Signal('15分钟_D1B_对称中枢_是_向下_7笔_0')

    :param c: CZSC对象
    :param di: 倒数第 di 笔
    :return: s
    """
    bis: List[BI] = get_sub_elements(c.bi_list, di=di, n=10)
    k1, k2, k3 = f"{c.freq.value}_D{di}B_对称中枢".split("_")
    for i in (7, 5, 3):
        v1 = is_symmetry_zs(bis[-i:])
        if v1:
            v3 = f"{i}笔"
            break
        else:
            v3 = "任意"

    v1 = "是" if v1 else "否"
    v2 = "向上" if bis[-1].direction == Direction.Down else "向下"

    s = OrderedDict()
    x1 = Signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2, v3=v3)
    s[x1.key] = x1.value
    return s






# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2022/11/7 19:29
describe:  cxt 代表 CZSC 形态信号
"""
import numpy as np
from loguru import logger
from typing import List
from czsc import CZSC
from czsc.traders.base import CzscSignals
from czsc.objects import FX, BI, Direction, ZS, Mark
from czsc.utils import get_sub_elements, create_single_signal, is_bis_up, is_bis_down
from collections import OrderedDict


def cxt_bi_base_V230228(c: CZSC, **kwargs) -> OrderedDict:
    """BI基础信号

    **信号逻辑：**

    1. 取最后一个笔，最后一笔向下，则当前笔向上，最后一笔向上，则当前笔向下；
    2. 根据延伸K线数量判断当前笔的状态，中继或转折。

    **信号列表：**

    - Signal('15分钟_D0BL9_V230228_向下_中继_任意_0')
    - Signal('15分钟_D0BL9_V230228_向上_转折_任意_0')
    - Signal('15分钟_D0BL9_V230228_向下_转折_任意_0')
    - Signal('15分钟_D0BL9_V230228_向上_中继_任意_0')

    :param c: CZSC对象
    :param kwargs:
    :return: 信号识别结果
    """
    bi_init_length = kwargs.get('bi_init_length', 9)  # 笔的初始延伸长度，即笔的延伸长度小于该值时，笔的状态为转折，否则为中继
    k1, k2, k3 = f"{c.freq.value}_D0BL{bi_init_length}_V230228".split('_')
    v1 = '其他'
    if len(c.bi_list) < 3:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    last_bi = c.bi_list[-1]
    assert last_bi.direction in [Direction.Up, Direction.Down]
    v1 = '向上' if last_bi.direction == Direction.Down else '向下'
    v2 = "中继" if len(c.bars_ubi) >= bi_init_length else "转折"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2)


def cxt_fx_power_V221107(c: CZSC, di: int = 1) -> OrderedDict:
    """倒数第di个分型的强弱

    **信号列表：**

    - Signal('15分钟_D1F_分型强弱_中顶_有中枢_任意_0')
    - Signal('15分钟_D1F_分型强弱_弱底_有中枢_任意_0')
    - Signal('15分钟_D1F_分型强弱_强顶_有中枢_任意_0')
    - Signal('15分钟_D1F_分型强弱_弱顶_有中枢_任意_0')
    - Signal('15分钟_D1F_分型强弱_强底_有中枢_任意_0')
    - Signal('15分钟_D1F_分型强弱_中底_有中枢_任意_0')
    - Signal('15分钟_D1F_分型强弱_强顶_无中枢_任意_0')
    - Signal('15分钟_D1F_分型强弱_中顶_无中枢_任意_0')
    - Signal('15分钟_D1F_分型强弱_弱底_无中枢_任意_0')
    - Signal('15分钟_D1F_分型强弱_中底_无中枢_任意_0')
    - Signal('15分钟_D1F_分型强弱_弱顶_无中枢_任意_0')
    - Signal('15分钟_D1F_分型强弱_强底_无中枢_任意_0')

    :param c: CZSC 对象
    :param di: 倒数第di个分型
    :return:
    """
    k1, k2, k3 = f"{c.freq.value}_D{di}F_分型强弱".split("_")
    last_fx: FX = c.fx_list[-di]
    v1 = f"{last_fx.power_str}{last_fx.mark.value[0]}"
    v2 = "有中枢" if last_fx.has_zs else "无中枢"
    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2)


def cxt_first_buy_V221126(c: CZSC, di=1) -> OrderedDict:
    """一买信号

    **信号列表：**

    - Signal('15分钟_D1B_BUY1_一买_5笔_任意_0')
    - Signal('15分钟_D1B_BUY1_一买_11笔_任意_0')
    - Signal('15分钟_D1B_BUY1_一买_7笔_任意_0')
    - Signal('15分钟_D1B_BUY1_一买_21笔_任意_0')
    - Signal('15分钟_D1B_BUY1_一买_17笔_任意_0')
    - Signal('15分钟_D1B_BUY1_一买_19笔_任意_0')
    - Signal('15分钟_D1B_BUY1_一买_9笔_任意_0')
    - Signal('15分钟_D1B_BUY1_一买_15笔_任意_0')
    - Signal('15分钟_D1B_BUY1_一买_13笔_任意_0')

    :param c: CZSC 对象
    :param di: CZSC 对象
    :return: 信号字典
    """
    def __check_first_buy(bis: List[BI]):
        """检查 bis 是否是一买的结束

        :param bis: 笔序列，按时间升序
        """
        res = {"match": False, "v1": "一买", "v2": f"{len(bis)}笔", 'v3': "任意"}
        if len(bis) % 2 != 1 or bis[-1].direction == Direction.Up or bis[0].direction != bis[-1].direction:
            return res

        if max([x.high for x in bis]) != bis[0].high or min([x.low for x in bis]) != bis[-1].low:
            return res

        # 检查背驰：获取向下突破的笔列表
        key_bis = []
        for i in range(0, len(bis) - 2, 2):
            if i == 0:
                key_bis.append(bis[i])
            else:
                b1, _, b3 = bis[i - 2:i + 1]
                if b3.low < b1.low:
                    key_bis.append(b3)

        # 检查背驰：最后一笔的 power_price，power_volume，length 同时满足背驰条件才算背驰
        bc_price = bis[-1].power_price < max(bis[-3].power_price, np.mean([x.power_price for x in key_bis]))
        bc_volume = bis[-1].power_volume < max(bis[-3].power_volume, np.mean([x.power_volume for x in key_bis]))
        bc_length = bis[-1].length < max(bis[-3].length, np.mean([x.length for x in key_bis]))

        if bc_price and (bc_volume or bc_length):
            res['match'] = True
        return res

    k1, k2, k3 = c.freq.value, f"D{di}B", "BUY1"
    v1, v2, v3 = "其他", '任意', '任意'

    for n in (21, 19, 17, 15, 13, 11, 9, 7, 5):
        _bis = get_sub_elements(c.bi_list, di=di, n=n)
        if len(_bis) != n:
            logger.warning('笔的数量不对')
            continue

        _res = __check_first_buy(_bis)
        if _res['match']:
            v1, v2, v3 = _res['v1'], _res['v2'], _res['v3']
            break

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2, v3=v3)


def cxt_first_sell_V221126(c: CZSC, di=1) -> OrderedDict:
    """一卖信号

    **信号列表：**

    - Signal('15分钟_D1B_SELL1_一卖_17笔_任意_0')
    - Signal('15分钟_D1B_SELL1_一卖_15笔_任意_0')
    - Signal('15分钟_D1B_SELL1_一卖_5笔_任意_0')
    - Signal('15分钟_D1B_SELL1_一卖_7笔_任意_0')
    - Signal('15分钟_D1B_SELL1_一卖_9笔_任意_0')
    - Signal('15分钟_D1B_SELL1_一卖_19笔_任意_0')
    - Signal('15分钟_D1B_SELL1_一卖_21笔_任意_0')
    - Signal('15分钟_D1B_SELL1_一卖_13笔_任意_0')
    - Signal('15分钟_D1B_SELL1_一卖_11笔_任意_0')

    :param c: CZSC 对象
    :param di: CZSC 对象
    :return: 信号字典
    """

    def __check_first_sell(bis: List[BI]):
        """检查 bis 是否是一卖的结束

        :param bis: 笔序列，按时间升序
        """
        res = {"match": False, "v1": "一卖", "v2": f"{len(bis)}笔", 'v3': "任意"}
        if len(bis) % 2 != 1 or bis[-1].direction == Direction.Down:
            return res

        if bis[0].direction != bis[-1].direction:
            return res

        max_high = max([x.high for x in bis])
        min_low = min([x.low for x in bis])

        if max_high != bis[-1].high or min_low != bis[0].low:
            return res

        # 检查背驰：获取向上突破的笔列表
        key_bis = []
        for i in range(0, len(bis) - 2, 2):
            if i == 0:
                key_bis.append(bis[i])
            else:
                b1, _, b3 = bis[i - 2:i + 1]
                if b3.high > b1.high:
                    key_bis.append(b3)

        # 检查背驰：最后一笔的 power_price，power_volume，length 同时满足背驰条件才算背驰
        bc_price = bis[-1].power_price < max(bis[-3].power_price, np.mean([x.power_price for x in key_bis]))
        bc_volume = bis[-1].power_volume < max(bis[-3].power_volume, np.mean([x.power_volume for x in key_bis]))
        bc_length = bis[-1].length < max(bis[-3].length, np.mean([x.length for x in key_bis]))

        if bc_price and (bc_volume or bc_length):
            res['match'] = True
        return res

    k1, k2, k3 = c.freq.value, f"D{di}B", "SELL1"
    v1, v2, v3 = "其他", '任意', '任意'

    for n in (21, 19, 17, 15, 13, 11, 9, 7, 5):
        _bis = get_sub_elements(c.bi_list, di=di, n=n)
        if len(_bis) != n:
            logger.warning('笔的数量不对，跳过')
            continue

        _res = __check_first_sell(_bis)
        if _res['match']:
            v1, v2, v3 = _res['v1'], _res['v2'], _res['v3']
            break

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2, v3=v3)


def cxt_bi_break_V221126(c: CZSC, di=1) -> OrderedDict:
    """向上笔突破回调不破信号

    **信号列表：**

    - Signal('15分钟_D1B_向上_突破_5笔_任意_0')
    - Signal('15分钟_D1B_向上_突破_7笔_任意_0')
    - Signal('15分钟_D1B_向上_突破_9笔_任意_0')

    :param c: CZSC 对象
    :param di: CZSC 对象
    :return: 信号字典
    """

    def __check(bis: List[BI]):
        res = {"match": False, "v1": "突破", "v2": f"{len(bis)}笔", 'v3': "任意"}
        if len(bis) % 2 != 1 or bis[-1].direction == Direction.Up or bis[0].direction != bis[-1].direction:
            return res

        # 获取向上突破的笔列表
        key_bis = []
        for i in range(0, len(bis) - 2, 2):
            if i == 0:
                key_bis.append(bis[i])
            else:
                b1, _, b3 = bis[i - 2:i + 1]
                if b3.high > b1.high:
                    key_bis.append(b3)

        # 检查：
        # 1. 当下笔的最低点在任一向上突破笔的高点上
        # 2. 当下笔的最低点离笔序列最低点的距离不超过向上突破笔列表均值的1.618倍
        tb_break = bis[-1].low > min([x.high for x in key_bis])
        tb_price = bis[-1].low < min([x.low for x in bis]) + 1.618 * np.mean([x.power_price for x in key_bis])
        if tb_break and tb_price:
            res['match'] = True
        return res

    k1, k2, k3 = c.freq.value, f"D{di}B", "向上"
    v1, v2, v3 = "其他", '任意', '任意'

    for n in (9, 7, 5):
        _bis = get_sub_elements(c.bi_list, di=di, n=n)
        if len(_bis) != n:
            logger.warning('笔的数量不对，跳过')
            continue

        _res = __check(_bis)
        if _res['match']:
            v1, v2, v3 = _res['v1'], _res['v2'], _res['v3']
            break

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2, v3=v3)


def cxt_sub_b3_V221212(cat: CzscSignals, freq='60分钟', sub_freq='15分钟', th=10) -> OrderedDict:
    """小级别突破大级别中枢形成三买，贡献者：魏永超

    **信号逻辑：**

    1. freq级别中产生笔中枢，最后一笔向上时，中枢由之前3笔构成；最后一笔向下时，中枢由最后3笔构成。
    2. sub_freq级别中出现向上笔超越大级别中枢最高点，且随后不回到大级别中枢区间的th%以内。

    **信号列表：**

    - Signal('60分钟_15分钟_3买回踩10_确认_任意_任意_0')

    :param cat:
    :param freq: 中枢所在的大级别
    :param sub_freq: 突破大级别中枢，回踩形成小级别类3买的小级别
    :param th: 小级别回落对大级别中枢区间的回踩程度，0表示回踩不进大级别中枢，10表示回踩不超过中枢区间的10%
    :return: 信号识别结果
    """
    k1, k2, k3 = f"{freq}_{sub_freq}_三买回踩{th}".split('_')

    c: CZSC = cat.kas[freq]
    sub_c: CZSC = cat.kas[sub_freq]

    v1 = "其他"
    if len(c.bi_list) > 13 and len(sub_c.bi_list) > 10:
        last_bi = c.bi_list[-1]
        if last_bi.direction == Direction.Down:
            zs = ZS(symbol=cat.symbol, bis=c.bi_list[-3:])
        else:
            zs = ZS(symbol=cat.symbol, bis=c.bi_list[-4:-1])

        min7 = min([x.low for x in c.bi_list[-7:]])
        # 中枢成立，且中枢最低点不是最后7笔的最低点，且最后7笔最低点同时也是最后13笔最低点（保证低点起来第一个中枢）
        if zs.zd < zs.zg and zs.dd > min7 == min([x.low for x in c.bi_list[-13:]]):
            last_sub_bi = sub_c.bi_list[-1]

            if last_sub_bi.direction == Direction.Down and last_sub_bi.high > zs.gg \
                    and last_sub_bi.low > zs.zg - (th / 100) * (zs.zg - zs.zd):
                v1 = "确认"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def cxt_zhong_shu_gong_zhen_V221221(cat: CzscSignals, freq1='日线', freq2='60分钟') -> OrderedDict:
    """大小级别中枢共振，类二买共振；贡献者：琅盎

    **信号逻辑：**

    1. 不区分上涨或下跌中枢
    2. 次级别中枢 DD 大于本级别中枢中轴
    3. 次级别向下笔出底分型开多；反之看空

    **信号列表：**

    - Signal('日线_60分钟_中枢共振_看多_任意_任意_0')
    - Signal('日线_60分钟_中枢共振_看空_任意_任意_0')

    :param cat:
    :param freq1:大级别周期
    :param freq2: 小级别周期
    :return: 信号识别结果
    """
    k1, k2, k3 = f"{freq1}_{freq2}_中枢共振".split('_')

    max_freq: CZSC = cat.kas[freq1]
    min_freq: CZSC = cat.kas[freq2]
    symbol = cat.symbol

    def __is_zs(_bis):
        _zs = ZS(symbol=symbol, bis=_bis)
        if _zs.zd < _zs.zg:
            return True
        else:
            return False

    v1 = "其他"
    if len(max_freq.bi_list) >= 5 and __is_zs(max_freq.bi_list[-3:]) \
            and len(min_freq.bi_list) >= 5 and __is_zs(min_freq.bi_list[-3:]):

        big_zs = ZS(symbol=symbol, bis=max_freq.bi_list[-3:])
        small_zs = ZS(symbol=symbol, bis=min_freq.bi_list[-3:])

        if small_zs.dd > big_zs.zz and min_freq.bi_list[-1].direction == Direction.Down:
            v1 = "看多"

        if small_zs.gg < big_zs.zz and min_freq.bi_list[-1].direction == Direction.Up:
            v1 = "看空"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def cxt_bi_end_V230222(c: CZSC, **kwargs) -> OrderedDict:
    """当前是最后笔的第几次新低底分型或新高顶分型，用于笔结束辅助

    **信号逻辑：**

    1. 取最后笔及未成笔的分型，
    2. 当前如果是顶分型，则看当前顶分型是否新高，是第几个新高
    2. 当前如果是底分型，则看当前底分型是否新低，是第几个新低

    **信号列表：**

    - Signal('15分钟_D1MO3_结束辅助_新低_第1次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新低_第2次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新高_第1次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新高_第2次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新低_第3次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新低_第4次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新高_第3次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新高_第4次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新高_第5次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新低_第5次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新低_第6次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新高_第6次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新高_第7次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新低_第7次_任意_0')

    :param c: CZSC对象
    :param kwargs:
    :return: 信号识别结果
    """
    max_overlap = int(kwargs.get('max_overlap', 3))
    k1, k2, k3 = f"{c.freq.value}_D1MO{max_overlap}_结束辅助".split('_')
    v1 = '其他'
    v2 = '其他'

    if not c.ubi_fxs:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2)

    # 为了只取最后一笔以来的分型，没有用底层fx_list
    fxs = []
    if c.bi_list:
        fxs.extend(c.bi_list[-1].fxs[1:])
    ubi_fxs = c.ubi_fxs
    for x in ubi_fxs:
        if not fxs or x.dt > fxs[-1].dt:
            fxs.append(x)

    # 出分型那刻出信号，或者分型和最后一根bar相差 max_overlap 根K线时间内
    if (fxs[-1].elements[-1].dt == c.bars_ubi[-1].dt) or (c.bars_raw[-1].id - fxs[-1].raw_bars[-1].id <= max_overlap):
        if fxs[-1].mark == Mark.G:
            up = [x for x in fxs if x.mark == Mark.G]
            high_max = float('-inf')
            cnt = 0
            for fx in up:
                if fx.high > high_max:
                    cnt += 1
                    high_max = fx.high
            if fxs[-1].high == high_max:
                v1 = '新高'
                v2 = cnt

        else:
            down = [x for x in fxs if x.mark == Mark.D]
            low_min = float('inf')
            cnt = 0
            for fx in down:
                if fx.low < low_min:
                    cnt += 1
                    low_min = fx.low
            if fxs[-1].low == low_min:
                v1 = '新低'
                v2 = cnt

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=f"第{v2}次")


def cxt_bi_end_V230224(c: CZSC, **kwargs):
    """量价配合的笔结束辅助

    **信号逻辑：**

    1. 向下笔结束：fx_b 内最低的那根K线下影大于上影的两倍，同时fx_b内的平均成交量小于当前笔的平均成交量的0.618
    2. 向上笔结束：fx_b 内最高的那根K线上影大于下影的两倍，同时fx_b内的平均成交量大于当前笔的平均成交量的2倍

    **信号列表：**

    - Signal('15分钟_D1MO3_笔结束V230224_看多_任意_任意_0')
    - Signal('15分钟_D1MO3_笔结束V230224_看空_任意_任意_0')

    :param c: CZSC 对象
    :return: 信号字典
    """
    max_overlap = int(kwargs.get('max_overlap', 3))
    k1, k2, k3 = f"{c.freq.value}_D1MO{max_overlap}_笔结束V230224".split('_')
    v1 = '其他'
    if len(c.bi_list) <= 3:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    last_bi = c.bi_list[-1]
    bi_bars = last_bi.raw_bars
    bi_vol_mean = np.mean([x.vol for x in bi_bars])
    fx_bars = last_bi.fx_b.raw_bars
    fx_vol_mean = np.mean([x.vol for x in fx_bars])

    bar1 = fx_bars[np.argmin([x.low for x in fx_bars])]
    bar2 = fx_bars[np.argmax([x.high for x in fx_bars])]

    if bar1.upper > bar1.lower * 2 and fx_vol_mean > bi_vol_mean * 2:
        v1 = '看空'
    elif 2 * bar2.upper < bar2.lower and fx_vol_mean < bi_vol_mean * 0.618:
        v1 = '看多'
    else:
        v1 = '其他'

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


class BXT:
    """缠论笔形态识别基础类"""

    def __init__(self, bis: List[BI]):
        self.bis = bis
        self.xt_map = {
            '标准趋势': self.aAbBc,
            '类趋势': self.abcde,
            'aAb式盘整': self.aAb,
            'aAbcd式盘整': self.aAbcd,
            'abcAd式盘整': self.abcAd,
            'ABC式盘整': self.ABC,
            'BS2': self.BS2,
            'BS3': self.BS3,
        }

    @staticmethod
    def is_aAbBc(bis):
        """标准趋势"""
        # res 定义返回值标准
        res = {'match': False, 'v1': "任意", 'v2': "任意", 'v3': "任意"}

        if len(bis) >= 11:
            bi1, bi2, bi3, bi4, bi5, bi6, bi7, bi8, bi9, bi10, bi11 = bis[-11:]
            max_high = max([x.high for x in bis[-11:]])
            min_low = min([x.low for x in bis[-11:]])

            # 十一笔（2~4构成中枢A，8~10构成中枢B）
            if bi11.direction == Direction.Down and max_high == bi1.high and min_low == bi11.low \
                    and min(bi2.high, bi4.high) > max(bi2.low, bi4.low) \
                    and is_bis_down([bi5, bi6, bi7]) \
                    and min(bi2.low, bi4.low) > max(bi8.high, bi10.high) \
                    and min(bi8.high, bi10.high) > max(bi8.low, bi10.low):
                res = {'match': True, 'v1': "向下", 'v2': "11笔", 'v3': "A3B3"}
                return res

            if bi11.direction == Direction.Up and max_high == bi11.high and min_low == bi1.low \
                    and min(bi2.high, bi4.high) > max(bi2.low, bi4.low) \
                    and is_bis_up([bi5, bi6, bi7]) \
                    and max(bi2.high, bi4.high) < min(bi8.low, bi10.low) \
                    and min(bi8.high, bi10.high) > max(bi8.low, bi10.low):
                res = {'match': True, 'v1': "向上", 'v2': "11笔", 'v3': "A3B3"}
                return res

            # 十一笔（2~4构成中枢A，6~10构成中枢B）
            if bi11.direction == Direction.Down and max_high == bi1.high and min_low == bi11.low \
                    and min(bi2.high, bi4.high) > max(bi2.low, bi4.low) \
                    and min(bi2.low, bi4.low) > max(bi6.high, bi8.high, bi10.high) \
                    and min(bi6.high, bi8.high, bi10.high) > max(bi6.low, bi8.low, bi10.low):
                res = {'match': True, 'v1': "向下", 'v2': "11笔", 'v3': "A3B5"}
                return res

            if bi11.direction == Direction.Up and max_high == bi11.high and min_low == bi1.low \
                    and min(bi2.high, bi4.high) > max(bi2.low, bi4.low) \
                    and max(bi2.high, bi4.high) < min(bi6.low, bi8.low, bi10.low) \
                    and min(bi6.high, bi8.high, bi10.high) > max(bi6.low, bi8.low, bi10.low):
                res = {'match': True, 'v1': "向上", 'v2': "11笔", 'v3': "A3B5"}
                return res

            # 十一笔（2~6构成中枢A，8~10构成中枢B）
            if bi11.direction == Direction.Down and max_high == bi1.high and min_low == bi11.low \
                    and min(bi2.high, bi4.high, bi6.high) > max(bi2.low, bi4.low, bi6.low) \
                    and min(bi2.low, bi4.low, bi6.low) > max(bi8.high, bi10.high) \
                    and min(bi8.high, bi10.high) > max(bi8.low, bi10.low):
                res = {'match': True, 'v1': "向下", 'v2': "11笔", 'v3': "A5B3"}
                return res

            if bi11.direction == Direction.Up and max_high == bi11.high and min_low == bi1.low \
                    and min(bi2.high, bi4.high, bi6.high) > max(bi2.low, bi4.low, bi6.low) \
                    and max(bi2.high, bi4.high, bi6.high) < min(bi8.low, bi10.low) \
                    and min(bi8.high, bi10.high) > max(bi8.low, bi10.low):
                res = {'match': True, 'v1': "向上", 'v2': "11笔", 'v3': "A5B3"}
                return res

        if len(bis) >= 9:
            bi1, bi2, bi3, bi4, bi5, bi6, bi7, bi8, bi9 = bis[-9:]
            max_high = max([x.high for x in bis[-9:]])
            min_low = min([x.low for x in bis[-9:]])

            # 九笔（2~4构成中枢A，6~8构成中枢B）
            if bi9.direction == Direction.Down and max_high == bi1.high and min_low == bi9.low \
                    and min(bi2.high, bi4.high) > max(bi2.low, bi4.low) \
                    and min(bi2.low, bi4.low) > max(bi6.high, bi8.high) \
                    and min(bi6.high, bi8.high) > max(bi6.low, bi8.low):
                res = {'match': True, 'v1': "向下", 'v2': "9笔", 'v3': "A3B3"}
                return res

            if bi9.direction == Direction.Up and max_high == bi9.high and min_low == bi1.low \
                    and min(bi2.high, bi4.high) > max(bi2.low, bi4.low) \
                    and max(bi2.high, bi4.high) < min(bi6.low, bi8.low) \
                    and min(bi6.high, bi8.high) > max(bi6.low, bi8.low):
                res = {'match': True, 'v1': "向上", 'v2': "9笔", 'v3': "A3B3"}
                return res

        return res

    @property
    def aAbBc(self):
        """标准趋势"""
        return self.is_aAbBc(self.bis)

    @staticmethod
    def is_abcde(bis):
        """类趋势"""
        # res 定义返回值标准
        res = {'match': False, 'v1': "任意", 'v2': "任意", 'v3': "任意"}
        if len(bis) >= 9:
            bi1, bi2, bi3, bi4, bi5, bi6, bi7, bi8, bi9 = bis[-9:]

            if bi9.direction == Direction.Down and is_bis_down(bis[-9:]) \
                    and bi2.low > bi4.high and bi4.low > bi6.high and bi6.low > bi8.high:
                res = {'match': True, 'v1': "向下", 'v2': "9笔", 'v3': "任意"}
                return res

            if bi9.direction == Direction.Up and is_bis_up(bis[-9:]) \
                    and bi8.low > bi6.high and bi6.low > bi4.high and bi4.low > bi2.high:
                res = {'match': True, 'v1': "向上", 'v2': "9笔", 'v3': "任意"}
                return res

        if len(bis) >= 7:
            bi1, bi2, bi3, bi4, bi5, bi6, bi7 = bis[-7:]

            if bi7.direction == Direction.Down and is_bis_down(bis[-7:]) \
                    and bi2.low > bi4.high and bi4.low > bi6.high:
                res = {'match': True, 'v1': "向下", 'v2': "7笔", 'v3': "任意"}
                return res

            if bi7.direction == Direction.Up and is_bis_up(bis[-7:]) \
                    and bi6.low > bi4.high and bi4.low > bi2.high:
                res = {'match': True, 'v1': "向上", 'v2': "7笔", 'v3': "任意"}
                return res

        if len(bis) >= 5:
            bi1, bi2, bi3, bi4, bi5 = bis[-5:]

            if bi5.direction == Direction.Down and is_bis_down(bis[-5:]) \
                    and bi2.low > bi4.high:
                res = {'match': True, 'v1': "向下", 'v2': "5笔", 'v3': "任意"}
                return res

            if bi5.direction == Direction.Up and is_bis_up(bis[-5:]) \
                    and bi4.low > bi2.high:
                res = {'match': True, 'v1': "向上", 'v2': "5笔", 'v3': "任意"}
                return res

        return res

    @property
    def abcde(self):
        """类趋势"""
        return self.is_abcde(self.bis)

    @staticmethod
    def is_aAb(bis):
        """aAb式盘整"""
        # res 定义返回值标准
        res = {'match': False, 'v1': "任意", 'v2': "任意", 'v3': "任意"}
        if len(bis) >= 9:
            bi1, bi2, bi3, bi4, bi5, bi6, bi7, bi8, bi9 = bis[-9:]
            max_high = max([x.high for x in bis[-9:]])
            min_low = min([x.low for x in bis[-9:]])

            if bi9.direction == Direction.Down and max_high == bi1.high and bi9.low == min_low \
                    and min(bi2.high, bi4.high, bi6.high, bi8.high) > max(bi2.low, bi4.low, bi6.low, bi8.low):
                res = {'match': True, 'v1': "向下", 'v2': "9笔", 'v3': "任意"}
                return res

            if bi9.direction == Direction.Up and max_high == bi9.high and bi1.low == min_low \
                    and min(bi2.high, bi4.high, bi6.high, bi8.high) > max(bi2.low, bi4.low, bi6.low, bi8.low):
                res = {'match': True, 'v1': "向上", 'v2': "9笔", 'v3': "任意"}
                return res

        if len(bis) >= 7:
            bi1, bi2, bi3, bi4, bi5, bi6, bi7 = bis[-7:]
            max_high = max([x.high for x in bis[-7:]])
            min_low = min([x.low for x in bis[-7:]])

            if bi7.direction == Direction.Down and max_high == bi1.high and bi7.low == min_low \
                    and min(bi2.high, bi4.high, bi6.high) > max(bi2.low, bi4.low, bi6.low):
                res = {'match': True, 'v1': "向下", 'v2': "7笔", 'v3': "任意"}
                return res

            if bi7.direction == Direction.Up and max_high == bi7.high and bi1.low == min_low \
                    and min(bi2.high, bi4.high, bi6.high) > max(bi2.low, bi4.low, bi6.low):
                res = {'match': True, 'v1': "向上", 'v2': "7笔", 'v3': "任意"}
                return res

        if len(bis) >= 5:
            bi1, bi2, bi3, bi4, bi5 = bis[-5:]
            max_high = max([x.high for x in bis[-5:]])
            min_low = min([x.low for x in bis[-5:]])

            if bi5.direction == Direction.Down and max_high == bi1.high and bi5.low == min_low \
                    and bi2.low < bi4.high:
                res = {'match': True, 'v1': "向下", 'v2': "5笔", 'v3': "任意"}
                return res

            if bi5.direction == Direction.Up and max_high == bi5.high and bi1.low == min_low \
                    and bi4.low < bi2.high:
                res = {'match': True, 'v1': "向上", 'v2': "5笔", 'v3': "任意"}
                return res

        return res

    @property
    def aAb(self):
        """aAb式盘整"""
        return self.is_aAb(self.bis)

    @staticmethod
    def is_aAbcd(bis):
        """aAbcd式盘整"""
        # res 定义返回值标准
        res = {'match': False, 'v1': "任意", 'v2': "任意", 'v3': "任意"}

        if len(bis) >= 11:
            bi1, bi2, bi3, bi4, bi5, bi6, bi7, bi8, bi9, bi10, bi11 = bis[-11:]
            max_high = max([x.high for x in bis[-11:]])
            min_low = min([x.low for x in bis[-11:]])

            gg = max(bi2.high, bi4.high, bi6.high, bi8.high)
            zg = min(bi2.high, bi4.high, bi6.high, bi8.high)
            zd = max(bi2.low, bi4.low, bi6.low, bi8.low)
            dd = min(bi2.low, bi4.low, bi6.low, bi8.low)

            if bi11.direction == Direction.Down and max_high == bi1.high and bi11.low == min_low \
                    and zg >= zd >= dd > bi10.high:
                res = {'match': True, 'v1': "向下", 'v2': "11笔", 'v3': "任意"}
                return res

            if bi11.direction == Direction.Up and max_high == bi11.high and bi1.low == min_low \
                    and bi10.low > gg >= zg >= zd:
                res = {'match': True, 'v1': "向上", 'v2': "11笔", 'v3': "任意"}
                return res

        if len(bis) >= 9:
            bi1, bi2, bi3, bi4, bi5, bi6, bi7, bi8, bi9 = bis[-9:]
            max_high = max([x.high for x in bis[-9:]])
            min_low = min([x.low for x in bis[-9:]])

            gg = max(bi2.high, bi4.high, bi6.high)
            zg = min(bi2.high, bi4.high, bi6.high)
            zd = max(bi2.low, bi4.low, bi6.low)
            dd = min(bi2.low, bi4.low, bi6.low)

            if bi9.direction == Direction.Down and max_high == bi1.high and bi9.low == min_low \
                    and zg >= zd >= dd > bi8.high:
                res = {'match': True, 'v1': "向下", 'v2': "9笔", 'v3': "任意"}
                return res

            if bi9.direction == Direction.Up and max_high == bi9.high and bi1.low == min_low \
                    and bi8.low > gg >= zg >= zd:
                res = {'match': True, 'v1': "向上", 'v2': "9笔", 'v3': "任意"}
                return res

        if len(bis) >= 7:
            bi1, bi2, bi3, bi4, bi5, bi6, bi7 = bis[-7:]
            max_high = max([x.high for x in bis[-7:]])
            min_low = min([x.low for x in bis[-7:]])

            gg = max(bi2.high, bi4.high)
            zg = min(bi2.high, bi4.high)
            zd = max(bi2.low, bi4.low)
            dd = min(bi2.low, bi4.low)

            if bi7.direction == Direction.Down and max_high == bi1.high and bi7.low == min_low \
                    and zg >= zd >= dd > bi6.high:
                res = {'match': True, 'v1': "向下", 'v2': "7笔", 'v3': "任意"}
                return res

            if bi7.direction == Direction.Up and max_high == bi7.high and bi1.low == min_low \
                    and bi6.low > gg >= zg >= zd:
                res = {'match': True, 'v1': "向上", 'v2': "7笔", 'v3': "任意"}
                return res

        return res

    @property
    def aAbcd(self):
        """aAbcd式盘整"""
        return self.is_aAbcd(self.bis)

    @staticmethod
    def is_abcAd(bis):
        """abcAd式盘整"""
        # res 定义返回值标准
        res = {'match': False, 'v1': "任意", 'v2': "任意", 'v3': "任意"}
        if len(bis) >= 11:
            bi1, bi2, bi3, bi4, bi5, bi6, bi7, bi8, bi9, bi10, bi11 = bis[-11:]
            max_high = max([x.high for x in bis[-11:]])
            min_low = min([x.low for x in bis[-11:]])

            gg = max(bi4.high, bi6.high, bi8.high, bi10.high)
            zg = min(bi4.high, bi6.high, bi8.high, bi10.high)
            zd = max(bi4.low, bi6.low, bi8.low, bi10.low)
            dd = min(bi4.low, bi6.low, bi8.low, bi10.low)

            if bi11.direction == Direction.Down and max_high == bi1.high and bi11.low == min_low \
                    and bi2.low > gg >= zg >= zd:
                res = {'match': True, 'v1': "向下", 'v2': "11笔", 'v3': "任意"}
                return res

            if bi11.direction == Direction.Up and max_high == bi11.high and bi1.low == min_low \
                    and zg >= zd >= dd > bi2.high:
                res = {'match': True, 'v1': "向上", 'v2': "11笔", 'v3': "任意"}
                return res

        if len(bis) >= 9:
            bi1, bi2, bi3, bi4, bi5, bi6, bi7, bi8, bi9 = bis[-9:]
            max_high = max([x.high for x in bis[-9:]])
            min_low = min([x.low for x in bis[-9:]])

            gg = max(bi4.high, bi6.high, bi8.high)
            zg = min(bi4.high, bi6.high, bi8.high)
            zd = max(bi4.low, bi6.low, bi8.low)
            dd = min(bi4.low, bi6.low, bi8.low)

            if bi9.direction == Direction.Down and max_high == bi1.high and bi9.low == min_low \
                    and bi2.low > gg >= zg >= zd:
                res = {'match': True, 'v1': "向下", 'v2': "9笔", 'v3': "任意"}
                return res

            if bi9.direction == Direction.Up and max_high == bi9.high and bi1.low == min_low \
                    and zg >= zd >= dd > bi2.high:
                res = {'match': True, 'v1': "向上", 'v2': "9笔", 'v3': "任意"}
                return res

        if len(bis) >= 7:
            bi1, bi2, bi3, bi4, bi5, bi6, bi7 = bis[-7:]
            max_high = max([x.high for x in bis[-7:]])
            min_low = min([x.low for x in bis[-7:]])

            gg = max(bi4.high, bi6.high)
            zg = min(bi4.high, bi6.high)
            zd = max(bi4.low, bi6.low)
            dd = min(bi4.low, bi6.low)

            if bi7.direction == Direction.Down and max_high == bi1.high and bi7.low == min_low \
                    and bi2.low > gg >= zg >= zd:
                res = {'match': True, 'v1': "向下", 'v2': "7笔", 'v3': "任意"}
                return res

            if bi7.direction == Direction.Up and max_high == bi7.high and bi1.low == min_low \
                    and zg >= zd >= dd > bi2.high:
                res = {'match': True, 'v1': "向上", 'v2': "7笔", 'v3': "任意"}
                return res

        return res

    @property
    def abcAd(self):
        """abcAd式盘整"""
        return self.is_abcAd(self.bis)

    @staticmethod
    def is_ABC(bis):
        """ABC式盘整"""
        # res 定义返回值标准
        res = {'match': False, 'v1': "任意", 'v2': "任意", 'v3': "任意"}

        if len(bis) >= 11:
            bi1, bi2, bi3, bi4, bi5, bi6, bi7, bi8, bi9, bi10, bi11 = bis[-11:]
            max_high = max([x.high for x in bis[-11:]])
            min_low = min([x.low for x in bis[-11:]])

            if bi11.direction == Direction.Down and max_high == bi1.high and bi11.low == min_low:
                # A3B5C3
                if is_bis_down([bi1, bi2, bi3]) and is_bis_down([bi9, bi10, bi11]):
                    res = {'match': True, 'v1': "向下", 'v2': "11笔", 'v3': "A3B5C3"}
                    return res

                # A5B3C3
                if is_bis_down([bi1, bi2, bi3, bi4, bi5]) and is_bis_down([bi9, bi10, bi11]):
                    res = {'match': True, 'v1': "向下", 'v2': "11笔", 'v3': "A5B3C3"}
                    return res

                # A3B3C5
                if is_bis_down([bi1, bi2, bi3]) and is_bis_down([bi7, bi8, bi9, bi10, bi11]):
                    res = {'match': True, 'v1': "向下", 'v2': "11笔", 'v3': "A3B3C5"}
                    return res

            if bi11.direction == Direction.Up and max_high == bi11.high and bi1.low == min_low:
                # A3B5C3
                if is_bis_up([bi1, bi2, bi3]) and is_bis_up([bi9, bi10, bi11]):
                    res = {'match': True, 'v1': "向上", 'v2': "11笔", 'v3': "A3B5C3"}
                    return res

                # A5B3C3
                if is_bis_up([bi1, bi2, bi3, bi4, bi5]) and is_bis_up([bi9, bi10, bi11]):
                    res = {'match': True, 'v1': "向上", 'v2': "11笔", 'v3': "A5B3C3"}
                    return res

                # A3B3C5
                if is_bis_up([bi1, bi2, bi3]) and is_bis_up([bi7, bi8, bi9, bi10, bi11]):
                    res = {'match': True, 'v1': "向上", 'v2': "11笔", 'v3': "A3B3C5"}
                    return res

        if len(bis) >= 9:
            bi1, bi2, bi3, bi4, bi5, bi6, bi7, bi8, bi9 = bis[-9:]
            max_high = max([x.high for x in bis[-9:]])
            min_low = min([x.low for x in bis[-9:]])

            if bi9.direction == Direction.Down and max_high == bi1.high and bi9.low == min_low \
                    and is_bis_down([bi1, bi2, bi3]) and is_bis_down([bi7, bi8, bi9]):
                res = {'match': True, 'v1': "向下", 'v2': "9笔", 'v3': "任意"}
                return res

            if bi9.direction == Direction.Up and max_high == bi9.high and bi1.low == min_low \
                    and is_bis_up([bi1, bi2, bi3]) and is_bis_up([bi7, bi8, bi9]):
                res = {'match': True, 'v1': "向上", 'v2': "9笔", 'v3': "任意"}
                return res

        return res

    @property
    def ABC(self):
        """ABC式盘整"""
        return self.is_ABC(self.bis)

    @staticmethod
    def is_BS2(bis):
        """BS2"""
        # res 定义返回值标准
        res = {'match': False, 'v1': "任意", 'v2': "任意", 'v3': "任意"}

        if len(bis) >= 9:
            bi1, bi2, bi3, bi4, bi5, bi6, bi7, bi8, bi9 = bis[-9:]
            gg = max([bi2.high, bi4.high])
            zg = min([bi2.high, bi4.high])
            zd = max([bi2.low, bi4.low])
            dd = min([bi2.low, bi4.low])

            if bi9.direction == Direction.Down and is_bis_down([bi1, bi2, bi3, bi4, bi5]):
                if gg > bi9.low >= zg:
                    res = {'match': True, 'v1': "向下", 'v2': "左5右4", 'v3': "24上沿"}
                    return res

                if zg > bi9.low >= zd:
                    res = {'match': True, 'v1': "向下", 'v2': "左5右4", 'v3': "24内部"}
                    return res

                if zd > bi9.low >= dd:
                    res = {'match': True, 'v1': "向下", 'v2': "左5右4", 'v3': "24下沿"}
                    return res

            if bi9.direction == Direction.Up and is_bis_up([bi1, bi2, bi3, bi4, bi5]):
                if gg > bi9.high >= zg:
                    res = {'match': True, 'v1': "向上", 'v2': "左5右4", 'v3': "24上沿"}
                    return res

                if zg > bi9.high >= zd:
                    res = {'match': True, 'v1': "向上", 'v2': "左5右4", 'v3': "24内部"}
                    return res

                if zd > bi9.high >= dd:
                    res = {'match': True, 'v1': "向上", 'v2': "左5右4", 'v3': "24下沿"}
                    return res

        if len(bis) >= 7:
            bi1, bi2, bi3, bi4, bi5, bi6, bi7 = bis[-7:]
            gg = max([bi2.high, bi4.high])
            zg = min([bi2.high, bi4.high])
            zd = max([bi2.low, bi4.low])
            dd = min([bi2.low, bi4.low])

            if bi7.direction == Direction.Down and is_bis_down([bi1, bi2, bi3, bi4, bi5]):
                if gg > bi7.low >= zg:
                    res = {'match': True, 'v1': "向下", 'v2': "左5右2", 'v3': "24上沿"}
                    return res

                if zg > bi7.low >= zd:
                    res = {'match': True, 'v1': "向下", 'v2': "左5右2", 'v3': "24内部"}
                    return res

                if zd > bi7.low >= dd:
                    res = {'match': True, 'v1': "向下", 'v2': "左5右2", 'v3': "24下沿"}
                    return res

            if bi7.direction == Direction.Up and is_bis_up([bi1, bi2, bi3, bi4, bi5]):
                if gg > bi7.high >= zg:
                    res = {'match': True, 'v1': "向上", 'v2': "左5右2", 'v3': "24上沿"}
                    return res

                if zg > bi7.high >= zd:
                    res = {'match': True, 'v1': "向上", 'v2': "左5右2", 'v3': "24内部"}
                    return res

                if zd > bi7.high >= dd:
                    res = {'match': True, 'v1': "向上", 'v2': "左5右2", 'v3': "24下沿"}
                    return res

        if len(bis) >= 5:
            bi1, bi2, bi3, bi4, bi5 = bis[-5:]
            if bi5.direction == Direction.Down and is_bis_down([bi1, bi2, bi3]):
                if bi2.high > bi5.low > bi2.low:
                    res = {'match': True, 'v1': "向下", 'v2': "左3右2", 'v3': "笔2内部"}
                    return res

                if bi5.high > bi3.high > bi5.low > bi3.low:
                    res = {'match': True, 'v1': "向下", 'v2': "左3右2", 'v3': "笔3内部"}
                    return res

            if bi5.direction == Direction.Up and is_bis_up([bi1, bi2, bi3]):
                if bi2.high > bi5.high > bi2.low:
                    res = {'match': True, 'v1': "向上", 'v2': "左3右2", 'v3': "笔2内部"}
                    return res

                if bi5.high < bi3.high and bi5.low < bi3.low:
                    res = {'match': True, 'v1': "向上", 'v2': "左3右2", 'v3': "笔3内部"}
                    return res

        return res

    @property
    def BS2(self):
        return self.is_BS2(self.bis)

    @staticmethod
    def is_BS3(bis):
        """BS3"""
        # res 定义返回值标准
        res = {'match': False, 'v1': "任意", 'v2': "任意", 'v3': "任意"}

        if len(bis) >= 9:
            bi1, bi2, bi3, bi4, bi5, bi6, bi7, bi8, bi9 = bis[-9:]
            gg = max([bi2.high, bi4.high])
            dd = min([bi2.low, bi4.low])

            if bi9.direction == Direction.Down and is_bis_down([bi1, bi2, bi3, bi4, bi5]):
                if bi7.low < gg < bi9.low:
                    res = {'match': True, 'v1': "向下", 'v2': "左5右4", 'v3': "24上沿"}
                    return res

            if bi9.direction == Direction.Up and is_bis_up([bi1, bi2, bi3, bi4, bi5]):
                if bi9.high < dd < bi7.high:
                    res = {'match': True, 'v1': "向上", 'v2': "左5右4", 'v3': "24下沿"}
                    return res

        if len(bis) >= 7:
            bi1, bi2, bi3, bi4, bi5, bi6, bi7 = bis[-7:]
            gg = max([bi2.high, bi4.high])
            dd = min([bi2.low, bi4.low])

            if bi7.direction == Direction.Down and is_bis_down([bi1, bi2, bi3, bi4, bi5]):
                if bi7.low > gg:
                    res = {'match': True, 'v1': "向下", 'v2': "左5右2", 'v3': "24上沿"}
                    return res

            if bi7.direction == Direction.Up and is_bis_up([bi1, bi2, bi3, bi4, bi5]):
                if bi7.high < dd:
                    res = {'match': True, 'v1': "向上", 'v2': "左5右2", 'v3': "24下沿"}
                    return res

        if len(bis) >= 5:
            bi1, bi2, bi3, bi4, bi5 = bis[-5:]
            if bi5.direction == Direction.Down and is_bis_down([bi1, bi2, bi3]):
                if bi5.low > max(bi1.high, bi3.high):
                    res = {'match': True, 'v1': "向下", 'v2': "左3右2", 'v3': "任意"}
                    return res

            if bi5.direction == Direction.Up and is_bis_up([bi1, bi2, bi3]):
                if bi5.high < min(bi1.low, bi3.low):
                    res = {'match': True, 'v1': "向上", 'v2': "左3右2", 'v3': "任意"}
                    return res
        return res

    @property
    def BS3(self):
        return self.is_BS3(self.bis)







# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/12/13 17:48
describe: 验证信号计算的准确性，仅适用于缠论笔相关的信号，
          技术指标构建的信号，用这个工具检查不是那么方便
"""
import sys

sys.path.insert(0, '..')
import os
import numpy as np
from loguru import logger
from collections import OrderedDict
from czsc.data.ts_cache import TsDataCache
from czsc import CZSC, Signal
from czsc.traders.base import CzscTrader, check_signals_acc
from czsc.signals.tas import update_ma_cache
from czsc.utils import get_sub_elements, create_single_signal
from czsc import signals

os.environ['czsc_verbose'] = '1'


def bar_bpm_V230227(c: CZSC, di=1, n: int = 20, th: int = 1000, **kwargs) -> OrderedDict:
    """以BP为单位的绝对动量

    **信号逻辑：**

    1. 以BP为单位的绝对动量，计算最近n根K线的涨幅，如果大于th，则为超强，否则为强势；
    2. 反之，如果小于-th，则为超弱，否则为弱势

    **信号列表：**

    - Signal('15分钟_D2N5T300_绝对动量V230227_弱势_任意_任意_0')
    - Signal('15分钟_D2N5T300_绝对动量V230227_强势_任意_任意_0')
    - Signal('15分钟_D2N5T300_绝对动量V230227_超强_任意_任意_0')
    - Signal('15分钟_D2N5T300_绝对动量V230227_超弱_任意_任意_0')

    :param c: CZSC对象
    :param di: 倒数第几根K线
    :param n: 连续多少根K线
    :param th: 超过多少bp
    :return: 信号识别结果
    """
    k1, k2, k3, v1 = str(c.freq.value), f"D{di}N{n}T{th}", "绝对动量V230227", "其他"
    _bars = get_sub_elements(c.bars_raw, di=di, n=n)
    if len(_bars) != n:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    bp = (_bars[-1].close / _bars[0].open - 1) * 10000
    if bp > 0:
        v1 = "超强" if bp > th else "强势"
    else:
        v1 = "超弱" if abs(bp) > th else "弱势"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def get_signals(cat: CzscTrader) -> OrderedDict:
    s = OrderedDict({"symbol": cat.symbol, "dt": cat.end_dt, "close": cat.latest_price})
    # 使用缓存来更新信号的方法
    s.update(bar_bpm_V230227(cat.kas['15分钟'], di=2, n=5, th=300))
    return s


if __name__ == '__main__':
    data_path = r'C:\ts_data'
    dc = TsDataCache(data_path, sdt='2010-01-01', edt='20211209')

    symbol = '000001.SZ'
    raw_bars = dc.pro_bar_minutes(ts_code=symbol, asset='E', freq='15min',
                                  sdt='20181101', edt='20210101', adj='qfq', raw_bar=True)

    check_signals_acc(raw_bars, get_signals)



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/11/7 21:07
"""
import pandas as pd
from tqdm import tqdm
from copy import deepcopy
from loguru import logger
from typing import List
from collections import OrderedDict
from czsc import signals
from czsc.traders.base import CzscSignals, BarGenerator, CzscTrader
from czsc.objects import Signal, Factor, Event, Operate, Position
from test.test_analyze import read_1min, read_daily


def test_object_position():
    bars = read_daily()
    bg = BarGenerator(base_freq='日线', freqs=['周线', '月线'])
    for bar in bars[:1000]:
        bg.update(bar)

    def __get_signals(cat) -> OrderedDict:
        s = OrderedDict({"symbol": cat.symbol, "dt": cat.end_dt, "close": cat.latest_price})
        s.update(signals.cxt_first_buy_V221126(cat.kas['日线'], di=1))
        s.update(signals.bxt.get_s_three_bi(cat.kas['日线'], di=1))
        return s

    opens = [
        Event(name='开多', operate=Operate.LO, factors=[
            Factor(name="站上SMA5", signals_all=[
                Signal("日线_D1B_BUY1_一买_任意_任意_0"),
            ])
        ]),
        Event(name='开空', operate=Operate.SO, factors=[
            Factor(name="跌破SMA5", signals_all=[
                Signal("日线_D1B_BUY1_一卖_任意_任意_0"),
            ])
        ]),
    ]

    # 没有出场条件的测试
    pos = Position(name="测试A", symbol=bg.symbol, opens=opens, exits=[], interval=0, timeout=20, stop_loss=300)

    cs = CzscSignals(deepcopy(bg), get_signals=__get_signals)
    for bar in bars[1000:]:
        cs.update_signals(bar)
        pos.update(cs.s)

    df = pd.DataFrame(pos.pairs)
    assert df.shape == (16, 11)
    assert len(cs.s) == 13

    exits = [
        Event(name='平多', operate=Operate.LE, factors=[
            Factor(name="跌破SMA5", signals_all=[
                Signal("日线_倒1笔_三笔形态_向上收敛_任意_任意_0"),
            ])
        ]),
        Event(name='平空', operate=Operate.SE, factors=[
            Factor(name="站上SMA5", signals_all=[
                Signal("日线_倒1笔_三笔形态_向下收敛_任意_任意_0"),
            ])
        ]),
    ]

    pos = Position(name="测试B", symbol=bg.symbol, opens=opens, exits=exits, interval=0, timeout=20, stop_loss=300)

    cs = CzscSignals(deepcopy(bg), get_signals=__get_signals)
    for bar in bars[1000:]:
        cs.update_signals(bar)
        pos.update(cs.s)

    df = pd.DataFrame(pos.pairs)
    assert df.shape == (21, 11)
    assert len(cs.s) == 13


def test_generate_czsc_signals():
    from czsc.traders.base import generate_czsc_signals

    bars = read_daily()

    def __get_signals(cat) -> OrderedDict:
        s = OrderedDict({"symbol": cat.symbol, "dt": cat.end_dt, "close": cat.latest_price})
        s.update(signals.bxt.get_s_three_bi(cat.kas['日线'], di=1))
        s.update(signals.bxt.get_s_three_bi(cat.kas['周线'], di=1))
        s.update(signals.bxt.get_s_three_bi(cat.kas['月线'], di=1))
        s.update(signals.cxt_first_buy_V221126(cat.kas['日线'], di=1))
        s.update(signals.cxt_first_buy_V221126(cat.kas['日线'], di=2))
        s.update(signals.cxt_first_sell_V221126(cat.kas['日线'], di=1))
        s.update(signals.cxt_first_sell_V221126(cat.kas['日线'], di=2))
        return s

    res = generate_czsc_signals(bars, get_signals=__get_signals, freqs=['周线', '月线'], sdt="20100101", init_n=500)
    res_df = generate_czsc_signals(bars, get_signals=__get_signals, freqs=['周线', '月线'],
                                   sdt="20100101", init_n=500, df=True)

    assert len(res) == len(res_df)


def test_czsc_trader():
    bars = read_daily()

    sdt = "20100101"
    init_n = 2000
    sdt = pd.to_datetime(sdt)
    bars_left = [x for x in bars if x.dt < sdt]
    if len(bars_left) <= init_n:
        bars_left = bars[:init_n]
        bars_right = bars[init_n:]
    else:
        bars_right = [x for x in bars if x.dt >= sdt]

    bg = BarGenerator(base_freq='日线', freqs=['周线', '月线'])
    for bar in bars_left:
        bg.update(bar)

    def __get_signals(cat) -> OrderedDict:
        s = OrderedDict({"symbol": cat.symbol, "dt": cat.end_dt, "close": cat.latest_price})
        s.update(signals.bxt.get_s_three_bi(cat.kas['日线'], di=1))
        s.update(signals.cxt_first_buy_V221126(cat.kas['日线'], di=1))
        s.update(signals.cxt_first_buy_V221126(cat.kas['日线'], di=2))
        s.update(signals.cxt_first_sell_V221126(cat.kas['日线'], di=1))
        s.update(signals.cxt_first_sell_V221126(cat.kas['日线'], di=2))
        return s

    def __create_sma5_pos():
        opens = [
            Event(name='开多', operate=Operate.LO, factors=[
                Factor(name="站上SMA5", signals_all=[
                    Signal("日线_D1B_BUY1_一买_任意_任意_0"),
                ])
            ]),
            Event(name='开空', operate=Operate.SO, factors=[
                Factor(name="跌破SMA5", signals_all=[
                    Signal("日线_D1B_BUY1_一卖_任意_任意_0"),
                ])
            ]),
        ]

        exits = [
            Event(name='平多', operate=Operate.LE, factors=[
                Factor(name="跌破SMA5", signals_all=[
                    Signal("日线_倒1笔_三笔形态_向上收敛_任意_任意_0"),
                ])
            ]),
            Event(name='平空', operate=Operate.SE, factors=[
                Factor(name="站上SMA5", signals_all=[
                    Signal("日线_倒1笔_三笔形态_向下收敛_任意_任意_0"),
                ])
            ]),
        ]

        pos = Position(symbol=bg.symbol, opens=opens, exits=exits, interval=0, timeout=20, stop_loss=100)

        return pos

    def __create_sma10_pos():
        opens = [
            Event(name='开多', operate=Operate.LO, factors=[
                Factor(name="站上SMA5", signals_all=[
                    Signal("日线_倒1笔_三笔形态_向下无背_任意_任意_0"),
                ])
            ]),
            Event(name='开空', operate=Operate.SO, factors=[
                Factor(name="跌破SMA5", signals_all=[
                    Signal("日线_倒1笔_三笔形态_向上无背_任意_任意_0"),
                ])
            ]),
        ]

        exits = [
            Event(name='平多', operate=Operate.LE, factors=[
                Factor(name="跌破SMA5", signals_all=[
                    Signal("日线_倒1笔_三笔形态_向上收敛_任意_任意_0"),
                ])
            ]),
            Event(name='平空', operate=Operate.SE, factors=[
                Factor(name="站上SMA5", signals_all=[
                    Signal("日线_倒1笔_三笔形态_向下收敛_任意_任意_0"),
                ])
            ]),
        ]

        pos = Position(symbol=bg.symbol, opens=opens, exits=exits, interval=0, timeout=20, stop_loss=100)
        return pos

    def __create_sma20_pos():
        opens = [
            Event(name='开多', operate=Operate.LO, factors=[
                Factor(name="站上SMA5", signals_all=[
                    Signal("日线_D2B_BUY1_一买_任意_任意_0"),
                ])
            ]),
            Event(name='开空', operate=Operate.SO, factors=[
                Factor(name="跌破SMA5", signals_all=[
                    Signal("日线_D2B_BUY1_一卖_任意_任意_0"),
                ])
            ]),
        ]

        exits = [
            Event(name='平多', operate=Operate.LE, factors=[
                Factor(name="跌破SMA5", signals_all=[
                    Signal("日线_倒1笔_三笔形态_向上收敛_任意_任意_0"),
                ])
            ]),
            Event(name='平空', operate=Operate.SE, factors=[
                Factor(name="站上SMA5", signals_all=[
                    Signal("日线_倒1笔_三笔形态_向下收敛_任意_任意_0"),
                ])
            ]),
        ]

        pos = Position(symbol=bg.symbol, opens=opens, exits=exits, interval=0, timeout=20, stop_loss=100)
        return pos

    # 通过 update 执行
    ct = CzscTrader(deepcopy(bg), get_signals=__get_signals,
                    positions=[__create_sma5_pos(), __create_sma10_pos(), __create_sma20_pos()])
    for bar in bars_right:
        ct.update(bar)
        for _pos in ct.positions:
            if _pos.pos_changed:
                assert _pos.operates[-1]['dt'] == _pos.end_dt
                print(_pos.name, _pos.operates[-1], _pos.end_dt, _pos.pos)
                assert ct.pos_changed
        # print(f"{bar.dt}: pos_seq = {[x.pos for x in ct.positions]}mean_pos = {ct.get_ensemble_pos('mean')}; vote_pos = {ct.get_ensemble_pos('vote')}; max_pos = {ct.get_ensemble_pos('max')}")

    assert list(ct.positions[0].dump(False).keys()) == ['symbol', 'name', 'opens', 'exits', 'interval', 'timeout', 'stop_loss', 'T0']
    assert list(ct.positions[0].dump(True).keys()) == ['symbol', 'name', 'opens', 'exits', 'interval', 'timeout', 'stop_loss', 'T0', 'pairs', 'holds']
    assert [x.pos for x in ct.positions] == [0, -1, 0]

    # 测试自定义仓位集成
    def _weighted_ensemble(positions: List[Position]):
        return 0.5 * positions[0].pos + 0.5 * positions[1].pos

    assert ct.get_ensemble_pos(_weighted_ensemble) == -0.5
    assert ct.get_ensemble_pos('vote') == -1
    assert ct.get_ensemble_pos('max') == 0
    assert ct.get_ensemble_pos('mean') == -0.3333333333333333

    # 通过 on_bar 执行
    ct1 = CzscTrader(deepcopy(bg), get_signals=__get_signals,
                     positions=[__create_sma5_pos(), __create_sma10_pos(), __create_sma20_pos()])
    for bar in bars_right:
        ct1.on_bar(bar)
        # print(ct1.s)
        print(f"{ct1.end_dt}: pos_seq = {[x.pos for x in ct1.positions]}mean_pos = {ct1.get_ensemble_pos('mean')}; vote_pos = {ct1.get_ensemble_pos('vote')}; max_pos = {ct1.get_ensemble_pos('max')}")

    assert [x.pos for x in ct1.positions] == [0, -1, 0]

    assert len(ct1.positions[0].pairs) == len(ct.positions[0].pairs)
    assert len(ct1.positions[1].pairs) == len(ct.positions[1].pairs)
    assert len(ct1.positions[2].pairs) == len(ct.positions[2].pairs)

    # 通过 on_sig 执行
    from czsc.traders.base import generate_czsc_signals
    res = generate_czsc_signals(bars, get_signals=__get_signals, freqs=['周线', '月线'], sdt=sdt, init_n=init_n)
    ct2 = CzscTrader(positions=[__create_sma5_pos(), __create_sma10_pos(), __create_sma20_pos()])
    for sig in res:
        ct2.on_sig(sig)
        # print(ct2.s)
        print(f"{ct2.end_dt}: pos_seq = {[x.pos for x in ct2.positions]}mean_pos = {ct2.get_ensemble_pos('mean')}; vote_pos = {ct2.get_ensemble_pos('vote')}; max_pos = {ct2.get_ensemble_pos('max')}")

    assert [x.pos for x in ct2.positions] == [0, -1, 0]

    assert len(ct1.positions[0].pairs) == len(ct2.positions[0].pairs)
    assert len(ct1.positions[1].pairs) == len(ct2.positions[1].pairs)
    assert len(ct1.positions[2].pairs) == len(ct2.positions[2].pairs)


# def test_czsc_signals():
#     bars = read_daily()
#     bg = BarGenerator(base_freq='日线', freqs=['周线', '月线'])
#     for bar in bars[:1000]:
#         bg.update(bar)
#
#     cs = CzscSignals(bg, get_signals=get_signals)
#     for bar in bars[1000:]:
#         cs.update_signals(bar)
#     assert len(cs.s) == 14




# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/8/25 17:43
describe: 成交量相关信号
"""
from loguru import logger
try:
    import talib as ta
except:
    logger.warning(f"ta-lib 没有正确安装，相关信号函数无法正常执行。"
                   f"请参考安装教程 https://blog.csdn.net/qaz2134560/article/details/98484091")
import numpy as np
from collections import OrderedDict
from czsc.analyze import CZSC
from czsc.utils.sig import get_sub_elements, create_single_signal


def update_vol_ma_cache(c: CZSC, ma_type: str, timeperiod: int, **kwargs):
    """更新均线缓存

    :param c: CZSC对象
    :param ma_type: 均线类型
    :param timeperiod: 计算周期
    :return:
    """
    ma_type_map = {
        'SMA': ta.MA_Type.SMA,
        'EMA': ta.MA_Type.EMA,
        'WMA': ta.MA_Type.WMA,
        'KAMA': ta.MA_Type.KAMA,
        'TEMA': ta.MA_Type.TEMA,
        'DEMA': ta.MA_Type.DEMA,
        'MAMA': ta.MA_Type.MAMA,
        'TRIMA': ta.MA_Type.TRIMA,
    }

    ma_type = ma_type.upper()
    assert ma_type in ma_type_map.keys(), f"{ma_type} 不是支持的均线类型，可选值：{list(ma_type_map.keys())}"
    cache_key = f"VOL#{ma_type.upper()}{timeperiod}"

    if c.bars_raw[-1].cache and c.bars_raw[-1].cache.get(cache_key, None):
        # 如果最后一根K线已经有对应的缓存，不执行更新
        return cache_key

    last_cache = dict(c.bars_raw[-2].cache) if c.bars_raw[-2].cache else dict()
    if cache_key not in last_cache.keys() or len(c.bars_raw) < timeperiod + 15:
        # 初始化缓存
        data = np.array([x.vol for x in c.bars_raw], dtype=np.float64)
        ma = ta.MA(data, timeperiod=timeperiod, matype=ma_type_map[ma_type.upper()])
        assert len(ma) == len(data)
        for i in range(len(data)):
            _c = dict(c.bars_raw[i].cache) if c.bars_raw[i].cache else dict()
            _c.update({cache_key: ma[i] if ma[i] else data[i]})
            c.bars_raw[i].cache = _c

    else:
        # 增量更新最近3个K线缓存
        data = np.array([x.vol for x in c.bars_raw[-timeperiod - 10:]], dtype=np.float64)
        ma = ta.MA(data, timeperiod=timeperiod, matype=ma_type_map[ma_type.upper()])
        for i in range(1, 4):
            _c = dict(c.bars_raw[-i].cache) if c.bars_raw[-i].cache else dict()
            _c.update({cache_key: ma[-i]})
            c.bars_raw[-i].cache = _c
    return cache_key


def vol_single_ma_V230214(c: CZSC, di: int = 1, **kwargs) -> OrderedDict:
    """成交量单均线信号

    **信号逻辑：**

    1. vol > ma，多头；反之，空头
    2. ma[-1] > ma[-2]，向上；反之，向下

    **信号列表：**

    - Signal('日线_D2_VOL#SMA5_多头_向上_任意_0')
    - Signal('日线_D2_VOL#SMA5_空头_向上_任意_0')
    - Signal('日线_D2_VOL#SMA5_空头_向下_任意_0')
    - Signal('日线_D2_VOL#SMA5_多头_向下_任意_0')

    :param c: CZSC对象
    :param di: 信号计算截止倒数第i根K线
    :return: 信号识别结果
    """
    ma_type = kwargs.get("ma_type", "SMA")          # ma_type: 均线类型，必须是 `ma_type_map` 中的 key
    timeperiod = kwargs.get("timeperiod", 5)        # timeperiod: 均线计算周期
    cache_key = update_vol_ma_cache(c, ma_type, timeperiod)

    k1, k2, k3 = f"{c.freq.value}_D{di}_{cache_key}".split('_')
    bars = get_sub_elements(c.bars_raw, di=di, n=3)
    v1 = "多头" if bars[-1].vol >= bars[-1].cache[cache_key] else "空头"
    v2 = "向上" if bars[-1].cache[cache_key] >= bars[-2].cache[cache_key] else "向下"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2)


def vol_double_ma_V230214(c: CZSC, di: int = 1, t1: int = 5, t2: int = 20, **kwargs) -> OrderedDict:
    """成交量双均线信号

    **信号逻辑：**

    1. 短均线在长均线上方，看多；反之，看空

    **信号列表：**

    - Signal('日线_D2VOL#SMA5_VOL#SMA20_看空_任意_任意_0')
    - Signal('日线_D2VOL#SMA5_VOL#SMA20_看多_任意_任意_0')

    :param c: CZSC对象
    :param di: 信号计算截止倒数第i根K线
    :param t1: 短线均线
    :param t2: 长线均线
    :param kwargs:
    :return: 信号识别结果
    """
    assert t2 > t1, "t2必须是长线均线，t1为短线均线"
    ma_type = kwargs.get("ma_type", "SMA")
    cache_key1 = update_vol_ma_cache(c, ma_type, t1)
    cache_key2 = update_vol_ma_cache(c, ma_type, t2)

    k1, k2, k3 = f"{c.freq.value}_D{di}{cache_key1}_{cache_key2}".split('_')
    bars = get_sub_elements(c.bars_raw, di=di, n=3)
    v1 = "看多" if bars[-1].cache[cache_key1] >= bars[-1].cache[cache_key2] else "看空"
    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)





# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/11/21 17:48
describe: 信号系统，注意：这里仅仅只是提供一些写信号的例子，用来做策略是不太行的
"""

from . import bxt

# ======================================================================================================================
# 以下是 0.9.1 开始的新标准下实现的信号函数，规范定义：
# 1. 前缀3个字符区分信号类别
# 2. 后缀 V221107 之类的标识同一个信号函数的版本
# ======================================================================================================================

from czsc.signals.cxt import (
    cxt_fx_power_V221107,
    cxt_first_buy_V221126,
    cxt_first_sell_V221126,
    cxt_bi_break_V221126,
    cxt_sub_b3_V221212,
    cxt_zhong_shu_gong_zhen_V221221,
    cxt_bi_end_V230222,
    cxt_bi_end_V230224,
    cxt_bi_base_V230228,
)


from czsc.signals.byi import (
    byi_symmetry_zs_V221107,
)

from czsc.signals.coo import (
    coo_td_V221110,
)

from czsc.signals.vol import (
    vol_single_ma_V230214,
    vol_double_ma_V230214,
)

from czsc.signals.bar import (
    bar_end_V221111,
    bar_operate_span_V221111,
    bar_zdt_V221110,
    bar_zdt_V221111,
    bar_cross_ps_V221112,
    bar_section_momentum_V221112,
    bar_vol_grow_V221112,
    bar_mean_amount_V221112,
    bar_zdf_V221203,
    bar_accelerate_V221110,
    bar_accelerate_V221118,
    bar_fang_liang_break_V221216,
    bar_fake_break_V230204,
    bar_single_V230214,
    bar_amount_acc_V230214,
    bar_big_solid_V230215,
    bar_vol_bs1_V230224,
    bar_reversal_V230227,
    bar_bpm_V230227,
)

from czsc.signals.jcc import (
    jcc_san_xing_xian_V221023,
    jcc_ten_mo_V221028,
    jcc_bai_san_bin_V221030,
    jcc_san_fa_V20221115,
    jcc_san_fa_V20221118,
    jcc_wu_yun_gai_ding_V221101,
    jcc_ci_tou_V221101,
    jcc_xing_xian_V221118,
    jcc_fen_shou_xian_V20221113,
    jcc_yun_xian_V221118,
    jcc_zhu_huo_xian_V221027,
    jcc_ping_tou_V221113,
    jcc_zhuo_yao_dai_xian_v221113,
    jcc_two_crow_V221108,
    jcc_three_crow_V221108,
    jcc_three_soldiers_V221030,
    jcc_szx_V221111,
    jcc_ta_xing_V221124,
    jcc_san_szx_V221122,
    jcc_shan_chun_V221121,
    jcc_fan_ji_xian_V221121,
    jcc_gap_yin_yang_V221121,
)


from czsc.signals.tas import (
    update_macd_cache,
    update_ma_cache,
    update_kdj_cache,
    update_boll_cache,
    update_rsi_cache,

    tas_macd_base_V221028,
    tas_macd_change_V221105,
    tas_macd_direct_V221106,
    tas_macd_power_V221108,
    tas_macd_xt_V221208,
    tas_macd_bc_V221201,
    tas_macd_first_bs_V221201,
    tas_macd_first_bs_V221216,
    tas_macd_second_bs_V221201,

    tas_ma_base_V221101,
    tas_ma_base_V221203,
    tas_ma_round_V221206,
    tas_double_ma_V221203,

    tas_boll_power_V221112,
    tas_boll_bc_V221118,

    tas_kdj_base_V221101,
    tas_kdj_evc_V221201,

    tas_double_rsi_V221203,
    tas_rsi_base_V230227,

    tas_first_bs_V230217,
    tas_second_bs_V230228,
    tas_second_bs_V230303,

    tas_hlma_V230301,
    tas_hlma_V230304,
)





# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/12/13 17:48
describe: 验证信号计算的准确性，仅适用于缠论笔相关的信号，
          技术指标构建的信号，用这个工具检查不是那么方便
"""
import sys

sys.path.insert(0, '..')
import os
import numpy as np
from loguru import logger
from collections import OrderedDict
from czsc.data.ts_cache import TsDataCache
from czsc import CZSC, Signal
from czsc.traders.base import CzscTrader, check_signals_acc
from czsc.signals.tas import update_ma_cache
from czsc.utils import get_sub_elements, create_single_signal
from czsc import signals
from czsc.signals import update_rsi_cache


os.environ['czsc_verbose'] = '1'


def tas_rsi_base_V230227(c: CZSC, di=1, n: int = 6, th: int = 20, **kwargs) -> OrderedDict:
    """RSI超买超卖信号

    **信号逻辑：**

    在正常情况下，RSI指标都会在30-70的区间内波动。当6日RSI超过80时，表示市场已经处于超买区间。6日RSI达到90以上时，
    表示市场已经严重超买，股价极有可能已经达到阶段顶点。这时投资者应该果断卖出。当6日RSI下降到20时，表示市场已经处于
    超卖区间。6日RSI一旦下降到10以下，则表示市场已经严重超卖，股价极有可能会止跌回升，是很好的买入信号。

    **信号列表：**

    - Signal('日线_D2T20_RSI6V230227_超卖_向下_任意_0')
    - Signal('日线_D2T20_RSI6V230227_超买_向上_任意_0')
    - Signal('日线_D2T20_RSI6V230227_超买_向下_任意_0')
    - Signal('日线_D2T20_RSI6V230227_超卖_向上_任意_0')

    :param c: CZSC对象
    :param di: 倒数第几根K线
    :param n: RSI的计算周期
    :param th: RSI阈值
    :return: 信号识别结果
    """
    cache_key = update_rsi_cache(c, timeperiod=n)
    k1, k2, k3, v1 = str(c.freq.value), f"D{di}T{th}", f"{cache_key}V230227", "其他"
    _bars = get_sub_elements(c.bars_raw, di=di, n=2)
    if len(_bars) != 2:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    rsi1 = _bars[-1].cache[cache_key]
    rsi2 = _bars[-2].cache[cache_key]

    if rsi1 <= th:
        v1 = "超卖"
    elif rsi1 >= 100 - th:
        v1 = "超买"
    else:
        v1 = "其他"

    v2 = "向上" if rsi1 >= rsi2 else "向下"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2)


def get_signals(cat: CzscTrader) -> OrderedDict:
    s = OrderedDict({"symbol": cat.symbol, "dt": cat.end_dt, "close": cat.latest_price})
    # 使用缓存来更新信号的方法
    s.update(tas_rsi_base_V230227(cat.kas['日线'], di=2, n=6, th=20))
    return s


if __name__ == '__main__':
    data_path = r'C:\ts_data'
    dc = TsDataCache(data_path, sdt='2010-01-01', edt='20211209')

    symbol = '000001.SZ'
    raw_bars = dc.pro_bar_minutes(ts_code=symbol, asset='E', freq='15min',
                                  sdt='20181101', edt='20210101', adj='qfq', raw_bar=True)

    check_signals_acc(raw_bars, get_signals)



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2022/10/21 19:56
describe: 择时交易策略样例
"""
from loguru import logger
from czsc.data import TsDataCache, get_symbols
from czsc import signals
from collections import OrderedDict
from czsc.objects import Event, Position
from czsc.strategies import CzscStrategyBase


logger.disable('czsc.signals.cxt')


# 定义择时交易策略，策略函数名称必须是 trader_strategy
# ----------------------------------------------------------------------------------------------------------------------
class CzscStrategySMA5(CzscStrategyBase):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @classmethod
    def get_signals(cls, cat) -> OrderedDict:
        s = OrderedDict({"symbol": cat.symbol, "dt": cat.end_dt, "close": cat.latest_price})
        s.update(signals.bar_operate_span_V221111(cat.kas['15分钟'], k1='全天', span=('0935', '1450')))
        s.update(signals.bar_operate_span_V221111(cat.kas['15分钟'], k1='临收盘', span=('1410', '1450')))
        s.update(signals.bar_operate_span_V221111(cat.kas['15分钟'], k1='下午', span=('1300', '1450')))
        s.update(signals.bar_operate_span_V221111(cat.kas['15分钟'], k1='上午', span=('0935', '1130')))
        s.update(signals.bar_zdt_V221111(cat, '15分钟', di=1))
        # s.update(signals.bar_end_V221111(cat.kas['5分钟'], '15分钟'))
        s.update(signals.bar_mean_amount_V221112(cat.kas['日线'], di=2, n=20, th1=2, th2=1000))

        s.update(signals.tas_ma_base_V221101(cat.kas["日线"], di=1, ma_type='SMA', timeperiod=5))
        s.update(signals.tas_ma_base_V221101(cat.kas["日线"], di=2, ma_type='SMA', timeperiod=5))
        s.update(signals.tas_ma_base_V221101(cat.kas["日线"], di=5, ma_type='SMA', timeperiod=5))

        s.update(signals.cxt_first_buy_V221126(cat.kas['30分钟'], di=1))
        return s

    @property
    def positions(self):
        return [
            self.create_pos_a(),
        ]

    @property
    def freqs(self):
        return ['日线', '30分钟', '15分钟']

    def create_pos_a(self):
        opens = [
            {'name': '开多',
             'operate': '开多',
             'signals_all': ['日线_D2K20B均额_2至1000千万_是_任意_任意_0'],
             'signals_any': [],
             'signals_not': ['15分钟_D1K_涨跌停_涨停_任意_任意_0'],
             'factors': [
                 {'name': '站上SMA5',
                  'signals_all': ['全天_0935_1450_是_任意_任意_0',
                                  '日线_D1K_SMA5_多头_任意_任意_0',
                                  '日线_D5K_SMA5_空头_向下_任意_0',
                                  '30分钟_D1B_BUY1_一买_任意_任意_0'],
                  'signals_any': [],
                  'signals_not': []}
             ]},
            {'name': '开空',
             'operate': '开空',
             'signals_all': ['日线_D2K20B均额_2至1000千万_是_任意_任意_0'],
             'signals_any': [],
             'signals_not': ['15分钟_D1K_涨跌停_跌停_任意_任意_0'],
             'factors': [
                 {'name': '站上SMA5',
                  'signals_all': ['全天_0935_1450_是_任意_任意_0',
                                  '日线_D1K_SMA5_空头_任意_任意_0',
                                  '日线_D5K_SMA5_多头_向下_任意_0',
                                  '30分钟_D1B_BUY1_一卖_任意_任意_0'],
                  'signals_any': [],
                  'signals_not': []}
             ]},
        ]

        exits = [
            {'name': '平多',
             'operate': '平多',
             'signals_all': [],
             'signals_any': [],
             'signals_not': ['15分钟_D1K_涨跌停_跌停_任意_任意_0'],
             'factors': [
                 {'name': '跌破SMA5',
                  'signals_all': ['下午_1300_1450_是_任意_任意_0',
                                  '日线_D1K_SMA5_空头_任意_任意_0',
                                  '日线_D2K_SMA5_多头_任意_任意_0'],
                  'signals_any': [],
                  'signals_not': []}
             ]}
        ]
        pos = Position(name="A", symbol=self.symbol,
                       opens=[Event.load(x) for x in opens],
                       exits=[Event.load(x) for x in exits],
                       interval=3600*4, timeout=100, stop_loss=1000, T0=False)
        return pos


# 定义命令行接口的特定参数
# ----------------------------------------------------------------------------------------------------------------------

# 【必须】执行结果路径
results_path = r"D:\ts_data\TS_SMA5"

# 初始化 Tushare 数据缓存
dc = TsDataCache(r"D:\ts_data")

# 【必须】策略回测参数设置
dummy_params = {
    "symbols": get_symbols(dc, 'train'),  # 回测使用的标的列表
    "sdt": "20150101",  # K线数据开始时间
    "mdt": "20200101",  # 策略回测开始时间
    "edt": "20220101",  # 策略回测结束时间
}


# 【可选】策略回放参数设置
replay_params = {
    "symbol": "000002.SZ#E",  # 回放交易品种
    "sdt": "20150101",  # K线数据开始时间
    "mdt": "20200101",  # 策略回放开始时间
    "edt": "20220101",  # 策略回放结束时间
}


# 【必须】定义K线数据读取函数，这里是为了方便接入任意数据源的K线行情
# ----------------------------------------------------------------------------------------------------------------------

def read_bars(symbol, sdt, edt):
    """自定义K线数据读取函数，便于接入任意来源的行情数据进行回测一类的分析

    :param symbol: 标的名称
    :param sdt: 行情开始时间
    :param edt: 行情介绍时间
    :return: list of RawBar
    """
    adj = 'hfq'
    freq = '15min'
    ts_code, asset = symbol.split("#")

    if "min" in freq:
        bars = dc.pro_bar_minutes(ts_code, sdt, edt, freq=freq, asset=asset, adj=adj, raw_bar=True)
    else:
        bars = dc.pro_bar(ts_code, sdt, edt, freq=freq, asset=asset, adj=adj, raw_bar=True)
    return bars





# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/12/13 17:48
describe: 验证信号计算的准确性，仅适用于缠论笔相关的信号，
          技术指标构建的信号，用这个工具检查不是那么方便
"""
import sys
sys.path.insert(0, '..')
import os
import numpy as np
from loguru import logger
from collections import OrderedDict
from czsc.data.ts_cache import TsDataCache
from czsc import CZSC, Signal
from czsc.objects import Mark
from czsc.traders.base import CzscTrader, check_signals_acc
from czsc.signals.tas import update_ma_cache
from czsc.utils import get_sub_elements, create_single_signal
from czsc import signals


os.environ['czsc_verbose'] = '1'

data_path = r'C:\ts_data'
dc = TsDataCache(data_path, sdt='2010-01-01', edt='20211209')

symbol = '000001.SZ'
bars = dc.pro_bar_minutes(ts_code=symbol, asset='E', freq='15min',
                          sdt='20181101', edt='20210101', adj='qfq', raw_bar=True)


def cxt_bi_end_V230222(c: CZSC, **kwargs) -> OrderedDict:
    """当前是最后笔的第几次新低底分型或新高顶分型，用于笔结束辅助

    **信号逻辑：**

    1. 取最后笔及未成笔的分型，
    2. 当前如果是顶分型，则看当前顶分型是否新高，是第几个新高
    2. 当前如果是底分型，则看当前底分型是否新低，是第几个新低

    **信号列表：**

    - Signal('15分钟_D1MO3_结束辅助_新低_第1次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新低_第2次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新高_第1次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新高_第2次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新低_第3次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新低_第4次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新高_第3次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新高_第4次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新高_第5次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新低_第5次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新低_第6次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新高_第6次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新高_第7次_任意_0')
    - Signal('15分钟_D1MO3_结束辅助_新低_第7次_任意_0')

    :param c: CZSC对象
    :param kwargs:
    :return: 信号识别结果
    """
    max_overlap = int(kwargs.get('max_overlap', 3))
    k1, k2, k3 = f"{c.freq.value}_D1MO{max_overlap}_结束辅助".split('_')
    v1 = '其他'
    v2 = '其他'

    if not c.ubi_fxs:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2)

    # 为了只取最后一笔以来的分型，没有用底层fx_list
    fxs = []
    if c.bi_list:
        fxs.extend(c.bi_list[-1].fxs[1:])
    ubi_fxs = c.ubi_fxs
    for x in ubi_fxs:
        if not fxs or x.dt > fxs[-1].dt:
            fxs.append(x)

    # 出分型那刻出信号，或者分型和最后一根bar相差 max_overlap 根K线时间内
    if (fxs[-1].elements[-1].dt == c.bars_ubi[-1].dt) or (c.bars_raw[-1].id - fxs[-1].raw_bars[-1].id <= max_overlap):
        if fxs[-1].mark == Mark.G:
            up = [x for x in fxs if x.mark == Mark.G]
            high_max = float('-inf')
            cnt = 0
            for fx in up:
                if fx.high > high_max:
                    cnt += 1
                    high_max = fx.high
            if fxs[-1].high == high_max:
                v1 = '新高'
                v2 = cnt

        else:
            down = [x for x in fxs if x.mark == Mark.D]
            low_min = float('inf')
            cnt = 0
            for fx in down:
                if fx.low < low_min:
                    cnt += 1
                    low_min = fx.low
            if fxs[-1].low == low_min:
                v1 = '新低'
                v2 = cnt

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=f"第{v2}次")


def get_signals(cat: CzscTrader) -> OrderedDict:
    s = OrderedDict({"symbol": cat.symbol, "dt": cat.end_dt, "close": cat.latest_price})
    # 使用缓存来更新信号的方法
    s.update(cxt_bi_end_V230222(cat.kas['15分钟']))
    return s


if __name__ == '__main__':
    check_signals_acc(bars, get_signals)










# coding: utf-8
import os

from . import qywx
from . import ta
from . import io
from . import echarts_plot

from .echarts_plot import kline_pro, heat_map
from .word_writer import WordWriter
from .corr import nmi_matrix, single_linear
from .bar_generator import BarGenerator, freq_end_time, resample_bars
from .io import dill_dump, dill_load, read_json, save_json
from .sig import check_pressure_support, check_gap_info, is_bis_down, is_bis_up, get_sub_elements
from .sig import same_dir_counts, fast_slow_cross, count_last_same, create_single_signal
from .plotly_plot import KlineChart


def x_round(x: [float, int], digit=4):
    """用去尾法截断小数

    :param x: 数字
    :param digit: 保留小数位数
    :return:
    """
    if isinstance(x, int):
        return x

    try:
        digit_ = pow(10, digit)
        x = int(x * digit_) / digit_
    except:
        print(f"x_round error: x = {x}")
    return x


def get_py_namespace(file_py: str, keys: list = None) -> dict:
    """获取 python 脚本文件中的 namespace

    :param file_py: python 脚本文件名
    :param keys: 指定需要的对象名称
    :return: namespace
    """
    text = open(file_py, 'r', encoding='utf-8').read()
    code = compile(text, file_py, 'exec')
    namespace = {"file_py": file_py, 'file_name': os.path.basename(file_py).split('.')[0]}
    exec(code, namespace)
    if keys:
        namespace = {k: v for k, v in namespace.items() if k in keys}
    return namespace


def freqs_sorted(freqs):
    """K线周期列表排序并去重，第一个元素是基础周期

    :param freqs: K线周期列表
    :return: K线周期排序列表
    """
    sf = ['Tick', '1分钟', '5分钟', '15分钟', '30分钟', '60分钟', '日线', '周线', '月线', '季线', '年线']
    _freqs_new = [x for x in sf if x in freqs]
    return _freqs_new

